@@ -44,6 +44,7 @@ static void                  gail_button_released_leave_handler (GtkWidget
 static void                  atk_action_interface_init  (AtkActionIface *iface);
 static gboolean              gail_button_do_action      (AtkAction      *action,
                                                          gint           i);
+static gboolean              idle_do_action             (gpointer       data);
 static gint                  gail_button_get_n_actions  (AtkAction      *action);
 static G_CONST_RETURN gchar* gail_button_get_description(AtkAction      *action,
                                                          gint           i);
@@ -73,6 +74,8 @@ static gboolean              gail_button_set_image_description
                                                          const gchar    *description);
 static GtkImage*             get_image_from_button      (GtkWidget      *button);
 static GtkWidget*            get_label_from_button      (GtkWidget      *button);
+static void                  set_role_for_button        (AtkObject      *accessible,
+                                                         GtkWidget      *button);
 
 static GailContainer* parent_class = NULL;
 
@@ -167,7 +170,7 @@ gail_button_new (GtkWidget *widget)
   accessible = ATK_OBJECT (object);
   atk_object_initialize (accessible, widget);
 
-  accessible->role = ATK_ROLE_PUSH_BUTTON;
+  set_role_for_button (accessible, widget);
 
   return accessible;
 }
@@ -285,8 +288,8 @@ static gboolean
 gail_button_do_action (AtkAction *action,
                        gint      i)
 {
-  GtkButton *button; 
   GtkWidget *widget;
+  GailButton *button;
   gboolean return_value = TRUE;
 
   widget = GTK_ACCESSIBLE (action)->widget;
@@ -296,11 +299,49 @@ gail_button_do_action (AtkAction *action,
      */
     return FALSE;
 
-  if (!GTK_WIDGET_SENSITIVE (widget) || !GTK_WIDGET_VISIBLE (widget))
+  if (!GTK_WIDGET_IS_SENSITIVE (widget) || !GTK_WIDGET_VISIBLE (widget))
+    return FALSE;
+
+  button = GAIL_BUTTON (action); 
+  if (button->action_idle_handler)
+    return FALSE;
+  switch (i)
+    {
+    case 0:
+    case 1:
+    case 2:
+      button->action_number = i;
+      button->action_idle_handler = gtk_idle_add (idle_do_action, button);
+      break;
+    default:
+      return_value = FALSE;
+      break;
+    }
+  return return_value; 
+}
+
+static gboolean
+idle_do_action (gpointer data)
+{
+  GtkButton *button; 
+  GtkWidget *widget;
+  GailButton *gail_button;
+
+  gail_button = GAIL_BUTTON (data);
+  gail_button->action_idle_handler = 0;
+
+  widget = GTK_ACCESSIBLE (gail_button)->widget;
+  if (widget == NULL)
+    /*
+     * State is defunct
+     */
+    return FALSE;
+
+  if (!GTK_WIDGET_IS_SENSITIVE (widget) || !GTK_WIDGET_VISIBLE (widget))
     return FALSE;
 
   button = GTK_BUTTON (widget); 
-  switch (i)
+  switch (gail_button->action_number)
     {
     case 0:
       gtk_widget_activate (widget);
@@ -331,10 +372,10 @@ gail_button_do_action (AtkAction *action,
       gtk_button_released (button);
       break;
     default:
-      return_value = FALSE;
+      g_assert_not_reached ();
       break;
     }
-  return return_value; 
+  return FALSE; 
 }
 
 static gint
@@ -588,7 +629,7 @@ get_image_from_button (GtkWidget *button)
   child = gtk_bin_get_child (GTK_BIN (button));
   if (GTK_IS_ALIGNMENT (child))
     child = gtk_bin_get_child (GTK_BIN (child));
-  if (GTK_IS_BOX (child))
+  if (GTK_IS_CONTAINER (child))
     {
       list = gtk_container_get_children (GTK_CONTAINER (child));
       if (!list)
@@ -734,6 +775,12 @@ gail_button_finalize (GObject            *object)
   g_free (button->press_description);
   g_free (button->release_description);
   g_free (button->click_keybinding);
+  if (button->action_idle_handler)
+    {
+      gtk_idle_remove (button->action_idle_handler);
+      button->action_idle_handler = 0;
+    }
+
   G_OBJECT_CLASS (parent_class)->finalize (object);
 }
 
@@ -745,10 +792,10 @@ get_label_from_button (GtkWidget *button)
   child = gtk_bin_get_child (GTK_BIN (button));
   if (GTK_IS_ALIGNMENT (child))
     child = gtk_bin_get_child (GTK_BIN (child));
-  if (GTK_IS_BOX (child))
+  while (GTK_IS_CONTAINER (child))
     {
       /*
-       * Child is not a label.  Search for a label in the box's children.
+       * Child is not a label.  Search for a label in the container's children.
        */
       GList *children, *tmp_list;
  
@@ -762,9 +809,46 @@ get_label_from_button (GtkWidget *button)
 	      child = GTK_WIDGET (tmp_list->data);
 	      break;
 	    }  
+           /*
+            * Label for button which are GtkreeView column headers are in a 
+            * GtkHBox in a GtkAlignment.
+            */
+          if (GTK_IS_ALIGNMENT (tmp_list->data))
+            {
+              child = gtk_bin_get_child (GTK_BIN (tmp_list->data));
+              break;
+	    }
 	}
 
       g_list_free (children);
     }
   return child;
 }
+
+static void
+set_role_for_button (AtkObject *accessible,
+                     GtkWidget *button)
+{
+  GtkWidget *parent;
+  AtkRole role;
+
+  parent = gtk_widget_get_parent (button);
+  if (GTK_IS_TREE_VIEW (parent))
+    {
+      role = ATK_ROLE_TABLE_COLUMN_HEADER;
+      /*
+       * Even though the accessible parent of the column header will
+       * be reported as the table because the parent widget of the
+       * GtkTreeViewColumn's button is the GtkTreeView we set
+       * the accessible parent for column header to be the table
+       * to ensure that atk_object_get_index_in_parent() returns
+       * the correct value; see gail_widget_get_index_in_parent().
+       */
+      atk_object_set_parent (accessible, gtk_widget_get_accessible (parent));
+    }
+  else
+    role = ATK_ROLE_PUSH_BUTTON;
+
+  accessible->role =  role;
+}
+
