@@ -51,6 +51,8 @@ struct _GailButton
   gchar         *press_description;
   gchar         *release_description;
   gchar         *click_keybinding;
+  guint         action_idle_handler;
+  gint          action_number;
 };
 
 GType gail_button_get_type (void);
