@@ -20,8 +20,7 @@
 #include <string.h>
 #include <gtk/gtk.h>
 #include "gaillabel.h"
-#include "gaillabelfactory.h"
-#include <util/gailmisc.h>
+#include <libgail-util/gailmisc.h>
 
 static void       gail_label_class_init            (GailLabelClass    *klass);
 static void	  gail_label_real_initialize	   (AtkObject 	      *obj,
@@ -37,7 +36,6 @@ static void       atk_text_interface_init          (AtkTextIface      *iface);
 static G_CONST_RETURN gchar* gail_label_get_name         (AtkObject         *accessible);
 static AtkStateSet*          gail_label_ref_state_set	 (AtkObject	    *accessible);
 static AtkRelationSet*       gail_label_ref_relation_set (AtkObject         *accessible);
-;
 
 /* atktext.h */
 
@@ -446,7 +444,7 @@ gail_label_get_text (AtkText *text,
   else
   {
     return gail_text_util_get_substring (GAIL_LABEL(text)->textutil, 
-                                         &start_pos, &end_pos);
+                                         start_pos, end_pos);
   }
 }
 
@@ -623,7 +621,7 @@ gail_label_get_selection (AtkText *text,
         return 0;
       else
         return gail_text_util_get_substring (GAIL_LABEL (text)->textutil, 
-                                             start_pos, end_pos);
+                                             *start_pos, *end_pos);
     }
   else 
     return NULL;
@@ -771,7 +769,12 @@ gail_label_get_offset_at_point (AtkText      *text,
                                               gtk_label_get_layout (label), 
                                               x_layout, y_layout, x, y, coords);
   if (index == -1)
-    return index;  
+    {
+      if (coords == ATK_XY_WINDOW || coords == ATK_XY_SCREEN)
+        return g_utf8_strlen (label->text, -1);
+
+      return index;  
+    }
   else
     return g_utf8_pointer_to_offset (label->text, label->text + index);  
 }
@@ -835,15 +838,6 @@ gail_label_get_default_attributes (AtkText        *text)
 
   label = GTK_LABEL (widget);
   
-  at_set = gail_misc_add_attribute (at_set, 
-                                    ATK_TEXT_ATTR_JUSTIFICATION, 
-     g_strdup (atk_text_attribute_get_value (ATK_TEXT_ATTR_JUSTIFICATION, 
-                                        gtk_label_get_justify (label))));
-  wrap_value = gtk_label_get_line_wrap (label);
-  at_set = gail_misc_add_attribute (at_set,
-                                    ATK_TEXT_ATTR_WRAP_MODE,
-              g_strdup (atk_text_attribute_get_value (ATK_TEXT_ATTR_WRAP_MODE, 
-                                                 wrap_value  ? 0 : 2))); 
   at_set = gail_misc_get_default_attributes (at_set,
                                              gtk_label_get_layout (label),
                                              widget);
@@ -866,6 +860,8 @@ gail_label_get_character_at_offset (AtkText	         *text,
 
   label = GTK_LABEL (widget);
   string = gtk_label_get_text (label);
+  if (offset >= g_utf8_strlen (string, -1))
+    return '\0';
   index = g_utf8_offset_to_pointer (string, offset);
 
   return g_utf8_get_char (index);
