@@ -52,6 +52,8 @@ static void     gail_widget_get_size       (AtkComponent    *component,
                                             gint            *width,
                                             gint            *height);
 
+static AtkLayer gail_widget_get_layer      (AtkComponent *component);
+
 static gboolean gail_widget_grab_focus     (AtkComponent    *component);
 
 
@@ -175,7 +177,6 @@ gail_widget_real_initialize (AtkObject *obj,
   g_return_if_fail (GTK_IS_WIDGET (data));
 
   widget = GTK_WIDGET (data);
-  obj->layer = ATK_LAYER_WIDGET;
 
   accessible = GTK_ACCESSIBLE (obj);
   accessible->widget = widget;
@@ -205,6 +206,8 @@ gail_widget_real_initialize (AtkObject *obj,
                     "unmap",
                     G_CALLBACK (gail_widget_map_gtk),
                     NULL);
+  g_object_set_data (G_OBJECT (obj), "atk-component-layer",
+		     GINT_TO_POINTER (ATK_LAYER_WIDGET));
 }
 
 AtkObject* 
@@ -244,7 +247,7 @@ gail_widget_connect_widget_destroyed (GtkAccessible *accessible)
 /*
  * This function is called when the widget is destroyed.
  * It sets the widget field in the GtkAccessible structure to NULL
- * and emits a tate-change signal for the state ATK_STATE_DEFUNCT
+ * and emits a state-change signal for the state ATK_STATE_DEFUNCT
  */
 static void 
 gail_widget_destroyed (GtkWidget     *widget,
@@ -252,7 +255,7 @@ gail_widget_destroyed (GtkWidget     *widget,
 {
   accessible->widget = NULL;
   atk_object_notify_state_change (ATK_OBJECT (accessible), ATK_STATE_DEFUNCT,
-                                  TRUE); 
+                                  TRUE);
 }
 
 static G_CONST_RETURN gchar*
@@ -454,7 +457,7 @@ gail_widget_get_index_in_parent (AtkObject *accessible)
     return -1;
   g_return_val_if_fail (GTK_IS_CONTAINER (parent_widget), -1);
 
-  children = gtk_container_children (GTK_CONTAINER (parent_widget));
+  children = gtk_container_get_children (GTK_CONTAINER (parent_widget));
 
   index = g_list_index (children, widget);
   g_list_free (children);
@@ -472,6 +475,7 @@ atk_component_interface_init (AtkComponentIface *iface)
   iface->add_focus_handler = gail_widget_add_focus_handler;
   iface->get_extents = gail_widget_get_extents;
   iface->get_size = gail_widget_get_size;
+  iface->get_layer = gail_widget_get_layer;
   iface->grab_focus = gail_widget_grab_focus;
   iface->remove_focus_handler = gail_widget_remove_focus_handler;
   iface->set_extents = gail_widget_set_extents;
@@ -491,7 +495,7 @@ gail_widget_add_focus_handler (AtkComponent    *component,
   signal_id = g_signal_lookup ("focus-event", ATK_TYPE_OBJECT);
 
   ret = g_signal_handler_find (component, match_type, signal_id, 0, NULL,
-                               handler, NULL);
+                               (gpointer) handler, NULL);
   if (!ret)
     {
       return g_signal_connect_closure_by_id (component, 
@@ -583,15 +587,28 @@ gail_widget_get_size (AtkComponent   *component,
   *height = widget->allocation.height;
 }
 
+static AtkLayer
+gail_widget_get_layer (AtkComponent *component)
+{
+  gint layer;
+  layer = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (component), "atk-component-layer"));
+
+  return (AtkLayer) layer;
+}
+
 static gboolean 
 gail_widget_grab_focus (AtkComponent   *component)
 {
   GtkWidget *widget = GTK_ACCESSIBLE (component)->widget;
+  GtkWidget *toplevel;
 
   g_return_val_if_fail (GTK_IS_WIDGET (widget), FALSE);
   if (GTK_WIDGET_CAN_FOCUS (widget))
     {
       gtk_widget_grab_focus (widget);
+      toplevel = gtk_widget_get_toplevel (widget);
+      if (GTK_WIDGET_TOPLEVEL (toplevel))
+        gtk_window_present (GTK_WINDOW (toplevel)); 
       return TRUE;
     }
   else
