@@ -49,6 +49,7 @@ static gboolean              gail_window_real_focus_gtk (GtkWidget     *widget,
 static AtkStateSet*          gail_window_ref_state_set  (AtkObject     *accessible);
 static void                  gail_window_real_notify_gtk (GObject      *obj,
                                                           GParamSpec   *pspec);
+static gint                  gail_window_get_mdi_zorder (AtkComponent  *component);
 
 static gboolean              gail_window_state_event_gtk (GtkWidget           *widget,
                                                           GdkEventWindowState *event);
@@ -129,7 +130,7 @@ gail_window_class_init (GailWindowClass *klass)
     g_signal_new ("activate",
                   G_TYPE_FROM_CLASS (klass),
                   G_SIGNAL_RUN_LAST,
-                  NULL, /* default signal handler */
+                  0, /* default signal handler */
                   NULL, NULL,
                   g_cclosure_marshal_VOID__VOID,
                   G_TYPE_NONE, 0);
@@ -137,7 +138,7 @@ gail_window_class_init (GailWindowClass *klass)
     g_signal_new ("create",
                   G_TYPE_FROM_CLASS (klass),
                   G_SIGNAL_RUN_LAST,
-                  NULL, /* default signal handler */
+                  0, /* default signal handler */
                   NULL, NULL,
                   g_cclosure_marshal_VOID__VOID,
                   G_TYPE_NONE, 0);
@@ -145,7 +146,7 @@ gail_window_class_init (GailWindowClass *klass)
     g_signal_new ("deactivate",
                   G_TYPE_FROM_CLASS (klass),
                   G_SIGNAL_RUN_LAST,
-                  NULL, /* default signal handler */
+                  0, /* default signal handler */
                   NULL, NULL,
                   g_cclosure_marshal_VOID__VOID,
                   G_TYPE_NONE, 0);
@@ -153,7 +154,7 @@ gail_window_class_init (GailWindowClass *klass)
     g_signal_new ("destroy",
                   G_TYPE_FROM_CLASS (klass),
                   G_SIGNAL_RUN_LAST,
-                  NULL, /* default signal handler */
+                  0, /* default signal handler */
                   NULL, NULL,
                   g_cclosure_marshal_VOID__VOID,
                   G_TYPE_NONE, 0);
@@ -161,7 +162,7 @@ gail_window_class_init (GailWindowClass *klass)
     g_signal_new ("maximize",
                   G_TYPE_FROM_CLASS (klass),
                   G_SIGNAL_RUN_LAST,
-                  NULL, /* default signal handler */
+                  0, /* default signal handler */
                   NULL, NULL,
                   g_cclosure_marshal_VOID__VOID,
                   G_TYPE_NONE, 0);
@@ -169,7 +170,7 @@ gail_window_class_init (GailWindowClass *klass)
     g_signal_new ("minimize",
                   G_TYPE_FROM_CLASS (klass),
                   G_SIGNAL_RUN_LAST,
-                  NULL, /* default signal handler */
+                  0, /* default signal handler */
                   NULL, NULL,
                   g_cclosure_marshal_VOID__VOID,
                   G_TYPE_NONE, 0);
@@ -177,7 +178,7 @@ gail_window_class_init (GailWindowClass *klass)
     g_signal_new ("resize",
                   G_TYPE_FROM_CLASS (klass),
                   G_SIGNAL_RUN_LAST,
-                  NULL, /* default signal handler */
+                  0, /* default signal handler */
                   NULL, NULL,
                   g_cclosure_marshal_VOID__VOID,
                   G_TYPE_NONE, 0);
@@ -185,7 +186,7 @@ gail_window_class_init (GailWindowClass *klass)
     g_signal_new ("restore",
                   G_TYPE_FROM_CLASS (klass),
                   G_SIGNAL_RUN_LAST,
-                  NULL, /* default signal handler */
+                  0, /* default signal handler */
                   NULL, NULL,
                   g_cclosure_marshal_VOID__VOID,
                   G_TYPE_NONE, 0);
@@ -239,6 +240,8 @@ gail_window_real_initialize (AtkObject *obj,
                     "window_state_event",
                     G_CALLBACK (gail_window_state_event_gtk),
                     NULL);
+  g_object_set_data (G_OBJECT (obj), "atk-component-layer",
+                     GINT_TO_POINTER (ATK_LAYER_WINDOW));
 }
 
 static G_CONST_RETURN gchar*
@@ -407,6 +410,7 @@ atk_component_interface_init (AtkComponentIface *iface)
 
   iface->get_extents = gail_window_get_extents;
   iface->get_size = gail_window_get_size;
+  iface->get_mdi_zorder = gail_window_get_mdi_zorder;
 }
 
 static void
@@ -488,3 +492,385 @@ gail_window_get_size (AtkComponent *component,
   *height = rect.height;
 }
 
+#include <X11/Xlib.h>
+#include <X11/Xatom.h>
+#include <gdk/gdkx.h>
+
+/* _NET_CLIENT_LIST_STACKING monitoring */
+
+typedef struct {
+  Window     *stacked_windows;
+  int         stacked_windows_len;
+  GdkWindow  *root_window;
+  guint       update_handler;
+  int        *desktop;
+  guint       update_desktop_handler;
+  gboolean   *desktop_changed;
+
+  guint       screen_initialized : 1;
+  guint       update_stacked_windows : 1;
+} GailScreenInfo;
+
+static GailScreenInfo *gail_screens = NULL;
+static int             num_screens = 0;
+static Atom            _net_client_list_stacking = None;
+static Atom            _net_wm_desktop = None;
+
+static gint
+get_window_desktop (Window window)
+{
+  Atom            ret_type;
+  int             format;
+  gulong          nitems;
+  gulong          bytes_after;
+  gulong         *cardinals;
+  int             error;
+  int             result;
+  int             desktop;
+
+  if (_net_wm_desktop == None)
+    _net_wm_desktop =
+		XInternAtom (gdk_display, "_NET_WM_DESKTOP", False);
+
+  gdk_error_trap_push ();
+  result = XGetWindowProperty (gdk_display, window, _net_wm_desktop,
+                               0, G_MAXLONG,
+                               False, XA_CARDINAL,
+                               &ret_type, &format, &nitems,
+                               &bytes_after, (guchar **) &cardinals);
+  error = gdk_error_trap_pop();
+  if (error != Success || result != Success)
+    return -1;
+
+  desktop = *cardinals;
+
+  XFree (cardinals);
+  if (nitems != 1)
+    return -1;
+  return desktop;
+}
+
+static void
+free_screen_info (GailScreenInfo *info)
+{
+  if (info->stacked_windows)
+    XFree (info->stacked_windows);
+  if (info->desktop)
+    g_free (info->desktop);
+  if (info->desktop_changed)
+    g_free (info->desktop_changed);
+
+  info->stacked_windows = NULL;
+  info->stacked_windows_len = 0;
+  info->desktop = NULL;
+  info->desktop_changed = NULL;
+}
+
+static gboolean
+get_stacked_windows (GailScreenInfo *info)
+{
+  Atom    ret_type;
+  int     format;
+  gulong  nitems;
+  gulong  bytes_after;
+  Window *data;
+  int     error;
+  int     result;
+  int     i;
+  int     j;
+  int    *desktops;
+  gboolean *desktops_changed;
+
+  if (_net_client_list_stacking == None)
+    _net_client_list_stacking =
+		XInternAtom (gdk_display, "_NET_CLIENT_LIST_STACKING", False);
+
+  gdk_error_trap_push ();
+  ret_type = None;
+  result = XGetWindowProperty (gdk_display,
+                               GDK_WINDOW_XWINDOW (info->root_window),
+                               _net_client_list_stacking,
+                               0, G_MAXLONG,
+                               False, XA_WINDOW, &ret_type, &format, &nitems,
+                               &bytes_after, (guchar **)&data);
+  error = gdk_error_trap_pop ();
+  if (error != Success || result != Success)
+    {
+      free_screen_info (info);
+      return FALSE;
+    }
+
+  if (ret_type != XA_WINDOW)
+    {
+      XFree (data);
+      free_screen_info (info);
+      return FALSE;
+    }
+
+  desktops = g_malloc0 (nitems * sizeof (int));
+  desktops_changed = g_malloc0 (nitems * sizeof (gboolean));
+  for (i = 0; i < nitems; i++)
+    {
+      gboolean window_found = FALSE;
+
+      for (j = 0; j < info->stacked_windows_len; j++)
+        {
+          if (info->stacked_windows [j] == data [i])
+            {
+              desktops [i] = info->desktop [j];
+              desktops_changed [i] = info->desktop_changed [j];
+              window_found = TRUE;
+              break;
+            }
+        }
+      if (!window_found)
+        {
+          desktops [i] = get_window_desktop (data [i]);
+          desktops_changed [i] = FALSE;
+        }
+    }
+  free_screen_info (info);
+  info->stacked_windows = data;
+  info->stacked_windows_len = nitems;
+  info->desktop = desktops;
+  info->desktop_changed = desktops_changed;
+
+  return TRUE;
+}
+
+static gboolean
+update_screen_info (gpointer data)
+{
+  int screen_n = GPOINTER_TO_INT (data);
+
+  gail_screens [screen_n].update_handler = 0;
+  gail_screens [screen_n].update_stacked_windows = FALSE;
+
+  get_stacked_windows (&gail_screens [screen_n]);
+
+  return FALSE;
+}
+
+static gboolean
+update_desktop_info (gpointer data)
+{
+  int screen_n = GPOINTER_TO_INT (data);
+  GailScreenInfo *info;
+  int i;
+
+  info = &gail_screens [screen_n];
+  info->update_desktop_handler = 0;
+
+  for (i = 0; i < info->stacked_windows_len; i++)
+    {
+      if (info->desktop_changed [i])
+        {
+          info->desktop [i] = get_window_desktop (info->stacked_windows [i]);
+          info->desktop_changed [i] = FALSE;
+        }
+    }
+
+  return FALSE;
+}
+
+static GdkFilterReturn
+filter_func (GdkXEvent *gdkxevent,
+	     GdkEvent  *event,
+	     gpointer   data)
+{
+  XEvent *xevent = gdkxevent;
+
+  if (xevent->type == PropertyNotify)
+    {
+      if (xevent->xproperty.atom == _net_client_list_stacking)
+        {
+          int     screen_n;
+
+          screen_n = gdk_screen_get_number (
+		gdk_drawable_get_screen (GDK_DRAWABLE (event->any.window)));
+
+          gail_screens [screen_n].update_stacked_windows = TRUE;
+          if (!gail_screens [screen_n].update_handler)
+            {
+              gail_screens [screen_n].update_handler = g_idle_add (update_screen_info,
+	        						   GINT_TO_POINTER (screen_n));
+            }
+        }
+      else if (xevent->xproperty.atom == _net_wm_desktop)
+        {
+          int     i;
+          int     j;
+          GailScreenInfo *info;
+
+          for (i = 0; i < num_screens; i++)
+            {
+              info = &gail_screens [i];
+              for (j = 0; j < info->stacked_windows_len; j++)
+                {
+                  if (xevent->xany.window == info->stacked_windows [j])
+                    {
+                      info->desktop_changed [j] = TRUE;
+                      if (!info->update_desktop_handler)
+                        {
+                          info->update_desktop_handler = g_idle_add (update_desktop_info, 
+                                                                     GINT_TO_POINTER (i));
+                        }
+                      break;
+                    }
+                }
+            }
+        }
+    }
+  return GDK_FILTER_CONTINUE;
+}
+
+static void
+display_closed (GdkDisplay *display,
+		gboolean    is_error)
+{
+  int i;
+
+  for (i = 0; i < num_screens; i++)
+    {
+      if (gail_screens [i].update_handler)
+	{
+	  g_source_remove (gail_screens [i].update_handler);
+	  gail_screens [i].update_handler = 0;
+	}
+
+      if (gail_screens [i].update_desktop_handler)
+	{
+	  g_source_remove (gail_screens [i].update_desktop_handler);
+	  gail_screens [i].update_desktop_handler = 0;
+	}
+
+      free_screen_info (&gail_screens [i]);
+    }
+
+  g_free (gail_screens);
+  gail_screens = NULL;
+  num_screens = 0;
+}
+
+static void
+init_gail_screens (void)
+{
+  GdkDisplay *display;
+
+  display = gdk_display_get_default ();
+
+  num_screens = gdk_display_get_n_screens (display);
+
+  gail_screens = g_new0 (GailScreenInfo, num_screens);
+  gdk_window_add_filter (NULL, filter_func, NULL);
+
+  g_signal_connect (display, "closed", G_CALLBACK (display_closed), NULL);
+}
+
+static void
+init_gail_screen (GdkScreen *screen,
+                  int        screen_n)
+{
+  XWindowAttributes attrs;
+
+  gail_screens [screen_n].root_window = gdk_screen_get_root_window (screen);
+
+  get_stacked_windows (&gail_screens [screen_n]);
+
+  XGetWindowAttributes (gdk_display,
+			GDK_WINDOW_XWINDOW (gail_screens [screen_n].root_window),
+			&attrs); 
+
+  XSelectInput (gdk_display,
+		GDK_WINDOW_XWINDOW (gail_screens [screen_n].root_window),
+		attrs.your_event_mask | PropertyChangeMask);
+           
+  gail_screens [screen_n].screen_initialized = TRUE;
+}
+
+static GailScreenInfo *
+get_screen_info (GdkScreen *screen)
+{
+  int screen_n;
+
+  g_return_val_if_fail (GDK_IS_SCREEN (screen), NULL);
+
+  screen_n = gdk_screen_get_number (screen);
+
+  if (gail_screens && gail_screens [screen_n].screen_initialized)
+    return &gail_screens [screen_n];
+
+  if (!gail_screens)
+    init_gail_screens ();
+
+  g_assert (gail_screens != NULL);
+
+  init_gail_screen (screen, screen_n);
+
+  g_assert (gail_screens [screen_n].screen_initialized);
+
+  return &gail_screens [screen_n];
+}
+
+static gint
+get_window_zorder (GdkWindow *window)
+{
+  GailScreenInfo *info;
+  Window          xid;
+  int             i;
+  int             zorder;
+  int             w_desktop;
+
+  g_return_val_if_fail (GDK_IS_WINDOW (window), -1);
+
+  info = get_screen_info (
+		gdk_drawable_get_screen (GDK_DRAWABLE (window)));
+
+  g_return_val_if_fail (info->stacked_windows != NULL, -1);
+
+  xid = GDK_WINDOW_XID (window);
+
+  w_desktop = -1;
+  for (i = 0; i < info->stacked_windows_len; i++)
+    {
+      if (info->stacked_windows [i] == xid)
+        {
+          w_desktop = info->desktop[i];
+          break;
+        }
+    }
+  if (w_desktop < 0)
+    return w_desktop;
+
+  zorder = 0;
+  for (i = 0; i < info->stacked_windows_len; i++)
+    {
+      if (info->stacked_windows [i] == xid)
+        {
+          return zorder;
+        }
+      else
+        {
+          if (info->desktop[i] == w_desktop)
+            zorder++;
+        }
+     }
+
+  return -1;
+}
+
+static gint
+gail_window_get_mdi_zorder (AtkComponent *component)
+{
+  GtkWidget *widget = GTK_ACCESSIBLE (component)->widget;
+
+  if (widget == NULL)
+    /*
+     * State is defunct
+     */
+    return -1;
+
+  g_return_val_if_fail (GTK_IS_WINDOW (widget), -1);
+
+  return get_window_zorder (widget->window);
+}
